{
  "cells": [
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "%matplotlib inline"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n\n# Pytorch and onnxruntime\n\nThis example compares the training between :epkg:`pytorch`\nand :epkg:`ORTModule`.\n\n## Functions\n\nThe first function creates the neural network.\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import time\nimport copy\nimport numpy\nimport onnx  # noqa\nfrom pandas import DataFrame\nfrom onnxruntime import get_device\nfrom sklearn.datasets import make_regression\nfrom sklearn.model_selection import train_test_split\nfrom sklearn.utils import shuffle\nfrom mlprodict.plotting.plotting_onnx import plot_onnx\nfrom tqdm import tqdm\nimport torch\nimport torch.nn.functional as F\nfrom onnxruntime.training import ORTModule\nfrom deeponnxcustom.tools.onnx_helper import save_as_onnx\n\n\ndef from_numpy(v, device=None, requires_grad=False):\n    \"\"\"\n    Convers a numpy array into a torch array and\n    sets *device* and *requires_grad*.\n    \"\"\"\n    v = torch.from_numpy(v)\n    if device is not None:\n        v = v.to(device)\n    v.requires_grad_(requires_grad)\n    return v\n\n\ndef build_class_model():\n\n    class CustomNet(torch.nn.Module):\n        def __init__(self, n_features, hidden_layer_sizes, n_output):\n            super(CustomNet, self).__init__()\n            self.hidden = []\n\n            size = n_features\n            for i, hid in enumerate(hidden_layer_sizes):\n                self.hidden.append(torch.nn.Linear(size, hid))\n                size = hid\n                setattr(self, \"hid%d\" % i, self.hidden[-1])\n            self.hidden.append(torch.nn.Linear(size, n_output))\n            setattr(self, \"predict\", self.hidden[-1])\n\n        def forward(self, x):\n            for hid in self.hidden:\n                x = hid(x)\n                x = F.relu(x)\n            return x\n\n    return CustomNet\n\n\ncls = build_class_model()\nmodel = cls(5, (4, 3), 1)\nsave_as_onnx(model, \"plot_ortmodule.onnx\", 5)\nplot_onnx(\"plot_ortmodule.onnx\", temp_dot=\"plot_ortmodule.dot\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Training\n\nThe function *train_model* works with a :class:`torch.nn.Module` and\n:epkg:`ORTModule`.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def train_model(nn, X_train, y_train, max_iter=25,\n                learning_rate_init=1e-5, batch_size=10,\n                device='cpu', opset=12, verbose=False,\n                use_ortmodule=False):\n    \"\"\"\n    Compares :epkg:`onnxruntime-training` to :epkg:`scikit-learn` for\n    training. Training algorithm is SGD.\n\n    :param nn: model to train\n    :param max_iter: number of iterations\n    :param learning_rate_init: initial learning rate\n    :param batch_size: batch size\n    :param device: `'cpu'` or `'cuda'`\n    :param opset: opset to choose for the conversion\n    :param use_ortmodule: use :epkg:`ORTModule`\n    :param verbose: displays intermediate information\n    \"\"\"\n    max_iter = int(max_iter)\n    learning_rate_init = float(learning_rate_init)\n    batch_size = int(batch_size)\n\n    if verbose:\n        print(\"N=%d\" % N)\n        print(\"n_features=%d\" % n_features)\n        print(\"hidden_layer_sizes=%r\" % (hidden_layer_sizes, ))\n        print(\"max_iter=%d\" % max_iter)\n        print(\"learning_rate_init=%f\" % learning_rate_init)\n        print(\"batch_size=%d\" % batch_size)\n        print(\"opset=%r (unused)\" % opset)\n        print(\"device=%r\" % device)\n\n    device0 = device\n    device = torch.device(\n        \"cuda:0\" if device in ('cuda', 'cuda:0', 'gpu') else \"cpu\")\n\n    if verbose:\n        print(\"fixed device=%r\" % device)\n        print('------------------')\n\n    if device0 == 'cpu':\n        nn.cpu()\n    else:\n        nn.cuda(device=device)\n\n    if verbose:\n        print(\"n_parameters=%d, n_layers=%d\" % (\n            len(list(nn.parameters())), len(nn.hidden)))\n        for i, p in enumerate(nn.parameters()):\n            print(\"  p[%d].shape=%r\" % (i, p.shape))\n\n    if use_ortmodule:\n        nn = ORTModule(nn)\n\n    optimizer = torch.optim.SGD(nn.parameters(), lr=learning_rate_init)\n    criterion = torch.nn.MSELoss(size_average=False)\n    batch_no = X_train.shape[0] // batch_size\n\n    # training\n    inputs = torch.tensor(\n        X_train[:1], requires_grad=True, device=device)\n    nn(inputs)\n\n    def train_torch():\n        if verbose:\n            loop = tqdm(range(max_iter))\n        else:\n            loop = range(max_iter)\n\n        losses = []\n        for epoch in loop:\n            running_loss = 0.0\n            x, y = shuffle(X_train, y_train)\n            for i in range(batch_no):\n                start = i * batch_size\n                end = start + batch_size\n                inputs = torch.tensor(\n                    x[start:end], requires_grad=True, device=device)\n                labels = torch.tensor(\n                    y[start:end], requires_grad=True, device=device)\n\n                def step_torch():\n                    optimizer.zero_grad()\n                    outputs = nn(inputs)\n                    loss = criterion(outputs, torch.unsqueeze(labels, dim=1))\n                    loss.backward()\n                    optimizer.step()\n                    return loss\n\n                loss = step_torch()\n                running_loss += loss.item()\n            losses.append(running_loss / X_train.shape[0])\n        return losses\n\n    begin = time.perf_counter()\n    losses = train_torch()\n    dur = time.perf_counter() - begin\n\n    if verbose:\n        print(\"time_torch=%r, running_loss=%r\" % (dur, losses[-1]))\n\n    return nn, losses, dur"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Some data.\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "N = 1000\nn_features = 5\n\nX, y = make_regression(N, n_features=n_features, bias=2)\nX = X.astype(numpy.float32)\ny = y.astype(numpy.float32)\nX_train, X_test, y_train, y_test = train_test_split(X, y)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "And one model\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "hidden_layer_sizes = (5, 3)\n\ncls_net = build_class_model()\nnn = cls_net(n_features, hidden_layer_sizes, 1)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Training with pytorch\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "device = \"cuda\" if get_device() == 'GPU' else 'cpu'\n\nt_x_train = from_numpy(X_train, device=device)\nt_y_train = from_numpy(y_train, device=device)\n\ntrained_nn_tch, losses_tch, duration = train_model(\n    copy.deepcopy(nn), t_x_train, t_y_train, verbose=True,\n    device=device, learning_rate_init=1e-4)\n\nprint(\"Torch, %d iterations, final loss=%f, duration=%f\" % (\n    len(losses_tch), losses_tch[-1], duration))\nprint('type(trained_nn_tch):', type(trained_nn_tch))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Training with ORTModule\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "trained_nn_ort, losses_ort, duration = train_model(\n    copy.deepcopy(nn), X_train, y_train, verbose=True,\n    device=device, learning_rate_init=1e-4, use_ortmodule=True)\n\nprint(\"ORT, %d iterations, final loss=%f, duration=%f\" % (\n    len(losses_ort), losses_ort[-1], duration))\nprint('type(trained_nn_ort):', type(trained_nn_ort))"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Visualisation\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "df = DataFrame(dict(torch=losses_tch, ort=losses_ort))\ndf.plot(title=\"Training loss / iterations\")"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## Performance\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "N = X_test.shape[0]\n\nt_x_test = from_numpy(X_test, device=device)\nt_y_test = from_numpy(y_test, device=device)\n\npred_tch = trained_nn_tch(t_x_test)\npred_ort = trained_nn_ort(t_x_test)\n\nerror_tch = ((pred_tch.ravel() - t_y_test.ravel()) ** 2).sum() / N\nerror_ort = ((pred_ort.ravel() - t_y_test.ravel()) ** 2).sum() / N\n\nprint(\"error torch: %f\" % error_tch)\nprint(\"error ort: %f\" % error_ort)\n\n# import matplotlib.pyplot as plt\n# plt.show()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.9.1"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}